#!/usr/bin/env python

import os
import unittest

# idmtools ...
from idmtools.assets import Asset
from idmtools.builders import SimulationBuilder
from idmtools.core.platform_factory import Platform
from idmtools.entities.experiment import Experiment

# emodpy
from emodpy.emod_task import EMODTask
# from emodpy.utils import EradicationBambooBuilds
# from emodpy.bamboo import get_model_files

# emod_api
import emod_api.campaign as camp
import emod_api.interventions.outbreak as ob
import emod_api.interventions.simple_vaccine as sv

import params
import manifest
"""
A test for Intervention Replacement Feature in DTK
Works with https://github.com/kfrey-idm/DtkTrunk/tree/kfrey_IVReplace02

DAD  = Dont_Allow_Duplicates
EIR  = Enable_Intervention_Replacement

Campaign built using the pattern:

Time       Event

10*N+1     Vaccine1            -  30% Acquision Blocking          (Expires after 6 days)
10*N+2     Vaccine2            -  60% Acquision Blocking          (Expires after 6 days)
10*N+3     OutbreakIndividual  - 100% Coverage                    (Infectious duration = 5 days; Infectivity = 0)
10*N+5     Examine Prevalence  - Ought to be ~70%, ~40%, or ~28%




N        Vaccine1            Vaccine2         DAD      EIR      Expected Prevalence

 0       Name = Default      Name = Default    0        0              28%
 1       Name = Default      Name = Default    0        1              28% # need to test this manually
 2       Name = Default      Name = Default    1        0              70%
 3       Name = Default      Name = Default    1        1              40%

 4       Name = Alice        Name = Default    0        0              28%
 5       Name = Alice        Name = Default    0        1              28% # need to test this manually
 6       Name = Alice        Name = Default    1        0              28%
 7       Name = Alice        Name = Default    1        1              28%

 8       Name = Alice        Name = Alice      0        0              28%
 9       Name = Alice        Name = Alice      0        1              28% # need to test this manually
10       Name = Alice        Name = Alice      1        0              70%
11       Name = Alice        Name = Alice      1        1              40%

12       Name = Alice        Name = Bob        0        0              28%
13       Name = Alice        Name = Bob        0        1              28% # need to test this manually
14       Name = Alice        Name = Bob        1        0              28%
15       Name = Alice        Name = Bob        1        1              28%
"""

implicit_config_set_fns = []


def update_sim_bic(simulation, value):
    simulation.task.config.parameters.Base_Infectivity_Constant = value
    return {"Base_Infectivity": value}


def update_sim_random_seed(simulation, value):
    simulation.task.config.parameters.Run_Number = value  # using from_default2
    return {"Run_Number": value}


def print_params():
    """
    Just a useful convenience function for the user.
    """
    # Display exp_name and nSims
    print("exp_name: ", params.exp_name)
    print("nSims: ", params.nSims)


def set_param_fn(config):
    config.parameters.Simulation_Type = 'GENERIC_SIM'
    config.parameters.Simulation_Duration = 160.0
    config.parameters.Base_Infectivity_Constant = 0
    config.parameters.Enable_Demographics_Reporting = 0 
    config.parameters.Incubation_Period_Constant = 0
    config.parameters.Infectious_Period_Constant = 5.0
    config.parameters.Enable_Vital_Dynamics = 0
    config.parameters.Enable_Disease_Mortality = 0
    config.parameters.Enable_Immune_Decay = 0
    config.parameters.Enable_Event_DB = 0
    config.parameters.Post_Infection_Acquisition_Multiplier = 1
    config.parameters.Post_Infection_Mortality_Multiplier = 1
    config.parameters.Post_Infection_Transmission_Multiplier = 1
    return config


def build_camp(params=None):
    print(f"Telling emod-api to use {manifest.schema_file} as schema.")
    camp.schema_path = manifest.schema_file
    sv.schema_path = manifest.schema_file

    box_duration = 6
    first = True
    ignore_immunity = False
    n = 0
    vaccine_efficacy_1 = 0.3
    vaccine_efficacy_2 = 0.6
    for sv_name_1, sv_name_2 in [('Vaccine', 'Vaccine'),
                                 ('Alice', 'Vaccine'),
                                 ('Alice', 'Alice'),
                                 ('Alice', 'Bob')]:
        for dont_allow_duplicates in [False, True]:
            for enable_intervention_replacement in [False, True]:
                event_1 = sv.new_intervention(timestep=10*n+1, v_type='Acquire', efficacy=vaccine_efficacy_1,
                                              sv_name=sv_name_1, waning_duration=box_duration)
                camp.add(event_1, first=first)
                if first:
                    first = False  # set it to False after the first event.

                event_2 = sv.new_intervention(timestep=10 * n + 2, v_type='Acquire', efficacy=vaccine_efficacy_2,
                                              sv_name=sv_name_2, waning_duration=box_duration,
                                              d_a_d=dont_allow_duplicates, e_i_r=enable_intervention_replacement)

                camp.add(event_2)

                # outbreak
                event_3 = ob.seed_by_coverage(camp, timestep=10*n+3, coverage=1,
                                              ignore_immunity=ignore_immunity)
                camp.add(event_3)
                n += 1
    return camp


def ep4_custom_cb(task):
    asset = Asset(os.path.join(manifest.ep4_path, manifest.my_ep4_assets), relative_path="python")
    task.common_assets.add_asset(asset)
    return task


def general_sim():
    print_params()

    task = EMODTask.from_default2(config_path="config.json", eradication_path=manifest.eradication_path,
                                  campaign_builder=build_camp, demog_builder=None,
                                  schema_path=manifest.schema_file, param_custom_cb=set_param_fn,
                                  ep4_custom_cb=ep4_custom_cb)

    print("Adding asset dir...")
    task.common_assets.add_directory(assets_directory=manifest.assets_input_dir)
    task.set_sif(manifest.sif)

    # Create simulation sweep with builder
    builder = SimulationBuilder()
    builder.add_sweep_definition(update_sim_random_seed, range(params.nSims))

    # create experiment from builder
    experiment = Experiment.from_builder(builder, task, name=params.exp_name)

    return experiment


class TestInterventionReplacement(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # todo: enable download from bamboo when the emod feature is merged into G-O branch.
        # right now we are using Eradication we built from Kurt's branch.
        # plan = EradicationBambooBuilds.GENERIC_LINUX
        # get_model_files(plan, manifest)
        pass

    def setUp(self) -> None:
        self.case_name = os.path.basename(__file__) + "--" + self._testMethodName
        print(self.case_name)

    def test_intervention_replacement(self):
        # Create a platform
        platform = Platform("SLURM")

        # Create an experiment
        exp = general_sim()
        # The last step is to call run() on the ExperimentManager to run the simulations.
        exp.run(wait_until_done=True, platform=platform)

        # Check result
        self.assertTrue(exp.succeeded, msg=f"Experiment {exp.uid} failed.\n")

        print(f"Experiment {exp.uid} succeeded.")

        for sim in exp.simulations:
            file = platform.get_files(sim, ["result.txt"])
            result = file["result.txt"].decode("utf-8")
            self.assertIn("Test result is True", result)


if __name__ == "__main__":
    unittest.main()
