import unittest
import json
from emod_api.campaign import *
import emod_api.demographics.DemographicsTemplates as DT
import emod_api.campaign as camp
import os
import manifest


import emodpy_generic
import emodpy_generic.demographics.GenericDemographics as Demographics
from emodpy_generic.interventions.polio_vaccine import *
from emodpy_generic.interventions.complex_import import *
from emodpy_generic.interventions.simple_sia import *
from emodpy_generic.interventions.emergence import *


from idmtools.assets import Asset, AssetCollection  #
from idmtools.builders import SimulationBuilder
from idmtools.core.platform_factory import Platform
from idmtools.entities.experiment import Experiment
from idmtools_platform_comps.utils.python_requirements_ac.requirements_to_asset_collection import RequirementsToAssetCollection
from idmtools_models.templated_script_task import get_script_wrapper_unix_task

# emodpy
from emodpy.emod_task import EMODTask
from emodpy.utils import EradicationBambooBuilds


class TestSimResults(unittest.TestCase):
    def setUp(self):
        current_directory = os.path.dirname(os.path.realpath(__file__))
        self.eradication_path = os.path.join(current_directory , "./inputs/bin/Eradication")
        self.schema_path = os.path.join(current_directory , "./inputs/bin/schema.json")

    def set_param_fn( self, config ):
        config.parameters.Simulation_Duration = 500.0
        config.parameters.Base_Infectivity_Constant = 3.5 
        config.parameters.Enable_Demographics_Reporting = 0 
        config.parameters.Incubation_Period_Constant = 0
        config.parameters.Infectious_Period_Exponential = 4.0 
        config.parameters.Minimum_End_Time = 500.0
        config.parameters.Enable_Report_Node_Event_Recorder = 1 # to check vaccine delay
        config.parameters.Acquisition_Blocking_Immunity_Decay_Rate = 0
        return config

    def set_param_fn_high_infectivity(self, config):
        config.parameters.Simulation_Duration = 500.0
        config.parameters.Base_Infectivity_Constant = 20
        config.parameters.Enable_Demographics_Reporting = 0 
        config.parameters.Incubation_Period_Constant = 0
        config.parameters.Infectious_Period_Exponential = 4.0
        config.parameters.Minimum_End_Time = 500.0
        #config.parameters.Report_Event_Recorder = 1 # to check vaccine delay
        #config.parameters.Report_Event_Recorder_Events = ['GP_EVENT_000', 'GP_EVENT_001', 'NewInfection']
        config.parameters.Enable_Report_Node_Event_Recorder = 1 # to check vaccine delay
        config.parameters.Enable_Immune_Decay = 0
        # config.parameters.Acquisition_Blocking_Immunity_Decay_Rate = 0
        # config.parameters.Transmission_Blocking_Immunity_Decay_Rate = 0
        return config

    def set_param_fn_no_infectivity(self, config):
        config.parameters.Simulation_Duration = 500.0
        config.parameters.Base_Infectivity_Constant = 0
        config.parameters.Enable_Demographics_Reporting = 0 
        config.parameters.Incubation_Period_Constant = 0
        config.parameters.Infectious_Period_Exponential = 4.0 
        config.parameters.Minimum_End_Time = 500
        config.parameters.Enable_Report_Node_Event_Recorder = 1 # to check vaccine delay
        config.parameters.Acquisition_Blocking_Immunity_Decay_Rate = 0
        return config

    
    def build_demog(self):

        # Have to use emod_api demographics here to avoid defaults
        demog = Demographics.fromBasicNode( lat=0, lon=0, pop=10000, name=1, forced_id=1)
        DT.EveryoneInitiallySusceptible(demog)
        return demog

    def update_sim_random_seed(self, simulation, value):
        simulation.task.config.parameters.Run_Number = value
        return {"Run_Number": value}

    # returns event coordinator and intervention config
    def parse_intervention(self, intervention):
        event_coord = intervention['Event_Coordinator_Config']
        intervention_coord = event_coord['Intervention_Config']
        return (event_coord, intervention_coord)

    def run_sim(self, build_camp, param_fn):
        self.experiment_name = self._testMethodName

        task = EMODTask.from_default2(config_path="config.json", eradication_path=self.eradication_path, campaign_builder=build_camp, 
                                        demog_builder=self.build_demog, schema_path=self.schema_path, param_custom_cb=param_fn, 
                                        ep4_custom_cb=None)
        builder = SimulationBuilder()
        builder.add_sweep_definition( self.update_sim_random_seed, range(1) )

        experiment  = Experiment.from_builder(builder, task, name=self.experiment_name) 
        self.platform = Platform("SLURM")
        task.set_sif(manifest.sif)
        experiment.run(wait_until_done=True, platform=self.platform)
        if not experiment.succeeded:
            print(f"Experiment {experiment.uid} failed.\n")
            exit()

        print(f"Experiment {experiment.uid} succeeded.")

        return experiment
    
    # Checks that basicEmergenceEvent imports people while basicSIAEvent tries to infect existing individuals
    def test_sia_event(self):
        def build_camp(params=None):

            camp.schema_path = self.schema_path
            
            emergence_event = basicSIAEvent(camp=camp, start_day=5)
            camp.add( emergence_event, first=True )

            return camp

        experiment = self.run_sim(build_camp, self.set_param_fn)

        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])

            channels = inset['Channels']
            new_infections = channels['New Infections']['Data']

            # emergence starts 3rd timestep, make sure this causes infection
            pre_infection = new_infections[0:3]
            post_infection = new_infections[4]

            self.assertTrue(all(num == 0 for num in pre_infection), msg="First 4 days should have no infections")
            self.assertGreater(post_infection, 0, msg="Emergence event should show individuals being infected on day 5")

    def test_basic_emergence_event(self):
        def build_camp(params=None):

            camp.schema_path = self.schema_path
            
            emergence_event = basicEmergenceEvent(camp=camp, start_day=7)
            camp.add( emergence_event, first=True )

            return camp

        experiment = self.run_sim(build_camp, self.set_param_fn)

        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])

            channels = inset['Channels']
            new_infections = channels['New Infections']['Data']
            population = channels['Statistical Population']['Data']

            pre_infection = new_infections[0:5]
            post_infection = new_infections[6]

            self.assertTrue(all(num == 0 for num in pre_infection), msg="First 4 days should have no infections")
            self.assertGreater(post_infection, 0, msg="Emergence event should show individuals being infected on day 5")
            self.assertGreater(population[6] - population[5], 15, msg="20 people should be imported on day 5 b")

    # Tests that SimpleSIA with 100% effectiveness and 100% coverage stops all new infections
    # Looks like effectiveness wanes
    def test_default_SIA(self):
        def build_camp(params=None):
            camp.schema_path = self.schema_path
            
            sia = SimpleSIA(camp=camp, timestep=2, Target_Age_Min=0, Target_Age_Max=100.0, Coverage=1.0)
            emergence_event = basicSIAEvent(camp=camp, start_day=1)
            camp.add(sia, first=True)
            camp.add( emergence_event )

            return camp

        experiment = self.run_sim(build_camp, self.set_param_fn)
        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])
            channels = inset['Channels']
            new_infections = channels['New Infections']['Data']
            population = channels['Statistical Population']['Data'][0]


            self.assertGreater(new_infections[0], 0.10*population, msg=f"Expected greater than zero infections on day 1, got 0")
            # Checks that after the SIA the maximum infections is equal to the difference between
            # population and the waning effect of the SIA
            for i in range(2, 365):
                self.assertLessEqual(new_infections[i], population - (population * 0.999**(i+1)), msg=f"Expected 0 infections on day 6, instead got {new_infections[5]}")

        
    # Tests that SimpleSIA with 100% effectiveness and 100% coverage stops all new infections
    def test_a_default_polio_vaccine1(self):
        def build_camp(params=None):

            camp.schema_path = self.schema_path
            
            # basic emergence event starts day 1
            emergence_event = basicSIAEvent(camp=camp, start_day=300)
            vaccine1 = SimpleIPV1(camp, 1, 1.0)

            camp.add( vaccine1, first=True)
            camp.add(emergence_event)

            return camp

        experiment = self.run_sim(build_camp, self.set_param_fn)
        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["stdout.txt", "output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])
            stdout = files["stdout.txt"].decode("utf-8")
            channels = inset['Channels']

            births = max(channels['Births']['Data'])
            number_vaccinated = channels['Births']['Data'][-1] - channels['Births']['Data'][270]  # DELAY AT 270 DAYS
            new_infections = sum(channels['New Infections']['Data'])
            min_population = channels['Statistical Population']['Data'][0]
            population = min_population + births  # total population is initial population + births

            # Checking that nobody is both vaccinated and infected
            self.assertLessEqual(number_vaccinated + new_infections, population, msg=f"Expected nobody could be both infected and vaccinated")
            self.assertAlmostEqual(number_vaccinated + new_infections, population, delta = 0.05*population, msg="Almost all people should be either vaccinated or infected")

    # Tests that 70% effective polio vaccine1 to everyone means no more than 30% get infected
    def test_mixed_polio_vaccine1(self):
        def build_camp(params=None):
            camp.schema_path = self.schema_path
            
            # basic emergence event starts day 1
            vaccine1 = SimpleIPV1(camp, 1, 0.7)
            emergence_event = basicSIAEvent(camp=camp, start_day=300)

            camp.add( vaccine1, first=True)
            camp.add(emergence_event)

            return camp
        experiment = self.run_sim(build_camp, self.set_param_fn)
        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])
            channels = inset['Channels']

            births = max(channels['Births']['Data'])
            number_vaccinated = (channels['Births']['Data'][-1] - channels['Births']['Data'][270]) * 0.7  # DELAY AT 270 DAYS
            new_infections = sum(channels['New Infections']['Data'])
            min_population = channels['Statistical Population']['Data'][0]
            population = min_population + births  # total population is initial population + birth

            # Population ~= new_infections + (births) * 1.3 since 30% of babies can also be infected
            self.assertLessEqual(new_infections + number_vaccinated, population)
            self.assertAlmostEqual(new_infections + number_vaccinated, population, delta=0.05*population)

    # Tests that vaccine2 with 100% effectiveness and 100% coverage stops all new infections
    def test_a_default_polio_vaccine2(self):
        def build_camp(params=None):

            camp.schema_path = self.schema_path
            
            emergence_event = basicSIAEvent(camp=camp, start_day=480)
            vaccine2 = SimpleIPV2(camp, 1, 1.0)

            camp.add( vaccine2, first=True )
            camp.add(emergence_event)

            return camp

        experiment = self.run_sim(build_camp, self.set_param_fn)
        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])
            channels = inset['Channels']

            births = max(channels['Births']['Data'])
            number_vaccinated = channels['Births']['Data'][-1] - channels['Births']['Data'][475]  # DELAY AT 475 DAYS
            new_infections = sum(channels['New Infections']['Data'])
            min_population = channels['Statistical Population']['Data'][0]
            population = min_population + births  # total population is initial population + birth

            # Checking that nobody is both vaccinated and infected
            self.assertLessEqual(number_vaccinated + new_infections, population, msg=f"Expected nobody could be both infected and vaccinated")
            self.assertAlmostEqual(number_vaccinated + new_infections, population, delta = 0.05*population, msg="Almost all people should be either vaccinated or infected")

    # Tests that 70% effective polio vaccine2 to everyone means no more than 30% get infected
    def test_mixed_polio_vaccine2(self):
        def build_camp(params=None):
            camp.schema_path = self.schema_path
            
            vaccine1 = SimpleIPV2(camp, 1, 1.0)
            emergence_event = basicSIAEvent(camp=camp, start_day=480)
            vaccine1['Event_Coordinator_Config']['Intervention_Config']['Actual_IndividualIntervention_Config']['Actual_IndividualIntervention_Configs'][0]['Acquire_Config']['Initial_Effect'] = 0.7

            camp.add( vaccine1, first=True )
            camp.add(emergence_event)

            return camp
        experiment = self.run_sim(build_camp, self.set_param_fn_high_infectivity)
        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])
            channels = inset['Channels']

            births = max(channels['Births']['Data'])
            number_vaccinated = (channels['Births']['Data'][-1] - channels['Births']['Data'][475]) * 0.7  # DELAY AT 475 DAYS
            new_infections = sum(channels['New Infections']['Data'])
            min_population = channels['Statistical Population']['Data'][0]
            population = min_population + births  # total population is initial population + birth

            # Population ~= new_infections + (births) * 1.3 since 30% of babies can also be infected
            self.assertLessEqual(new_infections + number_vaccinated, population)
            self.assertAlmostEqual(new_infections + number_vaccinated, population, delta=0.05*population)

    # Tests that complex import is seeding approximately the number of infected individuals as the daily import
    # pressure would indicate. Small buffer in assertion since the number in infected is picked from Poisson distribution
    def test_trickle_import(self):
        def build_camp(params=None):
            camp.schema_path = self.schema_path
            
            # basic import event starts day 1
            import_event = ComplexImportationEvent(camp=camp, dips=[1.0])
            
            camp.add( import_event, first=True )
            return camp
        experiment = self.run_sim(build_camp, self.set_param_fn_no_infectivity)
        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            inset = json.loads(files["output/InsetChart.json"])
            channels = inset['Channels']
            new_infections = channels['New Infections']['Data']
            self.assertAlmostEqual(sum(new_infections), 500, delta=50)

if __name__ == "__main__":
    unittest.main()
